@@ -3,7 +3,7 @@ typedload
 Module to load data into typed data structures
 """
 
-# Copyright (C) 2018-2020 Salvo "LtWorf" Tomaselli
+# Copyright (C) 2018-2021 Salvo "LtWorf" Tomaselli
 #
 # typedload is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -22,12 +22,13 @@ Module to load data into typed data structures
 
 
 import datetime
-from enum import Enum
+import ipaddress
 from pathlib import Path
 from typing import *
 
 from .exceptions import *
 from .typechecks import *
+from .helpers import tname
 
 
 __all__ = [
@@ -110,6 +111,10 @@ class Loader:
         handler. When disabled, the exceptions are not raised
         and the condition is considered False.
 
+    mangle_key: Defaults to 'name'
+        Specifies which key is used into the metadata dictionaries
+        to perform name-mangling.
+
     handlers: This is the list that the loader uses to
         perform its task.
         The type is:
@@ -127,6 +132,11 @@ class Loader:
         In most cases, it is sufficient to append new elements
         at the end, to handle more types.
 
+        There is an internal cache to speed up lookup, so after the
+        first call to load, this should no longer be modified.
+
+    strconstructed: Set of types to construct from a string.
+
     frefs: Dictionary to resolve ForwardRef.
         Something like
         class Node(NamedTuple):
@@ -154,26 +164,11 @@ class Loader:
     ones have any effect. This is to allow custom handlers to have their
     own parameters as well.
 
-    There is support for:
-        * Basic python types (int, str, bool, float, NoneType)
-        * NamedTuple
-        * Enum
-        * Optional[SomeType]
-        * List[SomeType]
-        * Dict[TypeA, TypeB]
-        * Tuple[TypeA, TypeB, TypeC]
-        * Tuple[SomeType, ...]
-        * Set[SomeType]
-        * Union[TypeA, TypeB]
-        * ForwardRef
-        * Literal
-        * Dataclass
-
     Using unions is complicated. If the types in the union are too
     similar to each other, it is easy to obtain an unexpected type.
     """
 
-    def __init__(self, **kwargs):
+    def __init__(self, **kwargs) -> None:
         # Types that do not need conversion
         self.basictypes = {int, bool, float, str, NONETYPE}
 
@@ -195,6 +190,20 @@ class Loader:
         # Enable conversion of dict-like things to dicts, before loading
         self.dictequivalence = True
 
+        # Which key is used in metadata to perform name mangling
+        self.mangle_key = 'name'
+
+        # Objects loaded from a string
+        self.strconstructed = {
+            Path,
+            ipaddress.IPv4Address,
+            ipaddress.IPv6Address,
+            ipaddress.IPv4Network,
+            ipaddress.IPv6Network,
+            ipaddress.IPv4Interface,
+            ipaddress.IPv6Interface
+        }
+
         # The list of handlers to use to load the data.
         # It gets iterated in order, and the first condition
         # that matches is used to load the value.
@@ -214,8 +223,9 @@ class Loader:
             (is_literal, _literalload),
             (is_typeddict, _namedtupleload),
             (lambda type_: type_ in {datetime.date, datetime.time, datetime.datetime}, _datetimeload),
-            (lambda type_: type_ == Path, _pathload),
+            (lambda type_: type_ in self.strconstructed, _strconstructload),
             (is_attrs, _attrload),
+            (is_any, _anyload),
         ]  # type: List[Tuple[Callable[[Any], bool], Callable[[Loader, Any, Type], Any]]]
 
         for k, v in kwargs.items():
@@ -247,6 +257,12 @@ class Loader:
 
         TypeError is raised if there is no known way to treat type_,
         otherwise all errors raise a ValueError.
+
+        annotation is used by recursive calls to track which path was
+        being executed in case of exceptions.
+
+        It is only needed when calling load recursively from
+        a custom handler.
         """
         p_index = self._indexcache.get(type_)
 
@@ -258,16 +274,16 @@ class Loader:
                 self._indexcache[type_] = index
             except ValueError:
                 raise TypedloadTypeError(
-                    'Cannot deal with value of type %s' % type_,
+                    'Cannot deal with value of type %s' % tname(type_),
                     value=value,
                     type_=type_
                 )
 
         # Add type to known types, to resolve ForwardRef later on
         if self.frefs is not None and hasattr(type_, '__name__'):
-            tname = type_.__name__
-            if tname not in self.frefs:
-                self.frefs[tname] = type_
+            typename = type_.__name__
+            if typename not in self.frefs:
+                self.frefs[typename] = type_
 
         func = self.handlers[index][1]
 
@@ -277,7 +293,7 @@ class Loader:
                 value = {k: v for k,v in value._get_kwargs()}
 
         try:
-            return cast(T, func(self, value, type_))
+            return func(self, value, type_)
         except Exception as e:
             assert isinstance(e, TypedloadException)
             e.trace.insert(0, TraceItem(value, type_, annotation))
@@ -304,6 +320,10 @@ def _forwardrefload(l: Loader, value: Any, type_: type) -> Any:
     return l.load(value, t, annotation=Annotation(AnnotationType.FORWARDREF, tname))
 
 
+def _anyload(l: Loader, value: Any, type_: type) -> Any:
+    return value
+
+
 def _literalload(l: Loader, value: Any, type_: type) -> Any:
     """
     Checks if the value is within the allowed literals and
@@ -311,7 +331,7 @@ def _literalload(l: Loader, value: Any, type_: type) -> Any:
     """
     if value in literalvalues(type_):
         return value
-    raise TypedloadValueError('Not one of the allowed values in %s' % type_, value=value, type_=type_)
+    raise TypedloadValueError('Not one of the allowed values in %s' % tname(type_), value=value, type_=type_)
 
 
 def _basicload(l: Loader, value: Any, type_: type) -> Any:
@@ -335,7 +355,7 @@ def _basicload(l: Loader, value: Any, type_: type) -> Any:
             except Exception as e:
                 raise TypedloadException(str(e), value=value, type_=type_)
         else:
-            raise TypedloadValueError('Not of type %s' % type_, value=value, type_=type_)
+            raise TypedloadValueError('Got %s of type %s, expected %s' % (repr(value), tname(type(value)), tname(type_)), value=value, type_=type_)
     return value
 
 
@@ -343,6 +363,8 @@ def _listload(l: Loader, value, type_) -> List:
     """
     This loads into something like List[int]
     """
+    if isinstance(value, dict):
+        raise TypedloadTypeError('Unable to load dictionary as a list', value=value, type_=type_)
     t = type_.__args__[0]
     try:
         return [l.load(v, t, annotation=Annotation(AnnotationType.INDEX, i)) for i, v in enumerate(value)]
@@ -371,6 +393,8 @@ def _setload(l: Loader, value, type_) -> Set:
     """
     This loads into something like Set[int]
     """
+    if isinstance(value, dict):
+        raise TypedloadTypeError('Unable to load dictionary as a set', value=value, type_=type_)
     t = type_.__args__[0]
     return {l.load(v, t, annotation=Annotation(AnnotationType.INDEX, i)) for i, v in enumerate(value)}
 
@@ -379,6 +403,8 @@ def _frozensetload(l: Loader, value, type_) -> FrozenSet:
     """
     This loads into something like FrozenSet[int]
     """
+    if isinstance(value, dict):
+        raise TypedloadTypeError('Unable to load dictionary as a frozenset', value=value, type_=type_)
     t = type_.__args__[0]
     return frozenset(l.load(v, t, annotation=Annotation(AnnotationType.INDEX, i)) for i, v in enumerate(value))
 
@@ -387,6 +413,8 @@ def _tupleload(l: Loader, value, type_) -> Tuple:
     """
     This loads into something like Tuple[int,str]
     """
+    if isinstance(value, dict):
+        raise TypedloadTypeError('Unable to load dictionary as a tuple', value=value, type_=type_)
     if HAS_TUPLEARGS:
         args = type_.__args__
     else:
@@ -396,13 +424,40 @@ def _tupleload(l: Loader, value, type_) -> Tuple:
         return tuple(l.load(v, args[0], annotation=Annotation(AnnotationType.INDEX, i)) for i, v in enumerate(value))
     else: # Tuple[something, something, somethingelse]
         if l.failonextra and len(value) > len(args):
-            raise TypedloadValueError('Value is too long for type %s' % type_, value=value, type_=type_)
+            raise TypedloadValueError('Value is too long for type %s' % tname(type_), value=value, type_=type_)
         elif len(value) < len(args):
-            raise TypedloadValueError('Value is too short for type %s' % type_, value=value, type_=type_)
+            raise TypedloadValueError('Value is too short for type %s' % tname(type_), value=value, type_=type_)
         return tuple(l.load(v, t, annotation=Annotation(AnnotationType.INDEX, i)) for i, (v, t) in enumerate(zip(value, args)))
 
 
-def _namedtupleload(l: Loader, value: Dict[str, Any], type_) -> Tuple:
+def _mangle_names(namesmap: Dict[str, str], value: Dict[str, Any], failonextra: bool) -> Dict[str, Any]:
+    """
+    Mangling names of a dictionary.
+
+    The dictionary is copied internally.
+
+    Namesmap is the mapping to be applied, in the format [dataname] = pyname
+    """
+    if not namesmap:
+        return value
+    r = {}
+
+    # Disallow the python names if they are in the map
+    skip = set(namesmap.values())
+
+    for k, v in value.items():
+        if k in skip and k not in namesmap:
+            if failonextra:
+                raise ValueError('Extra field: %s' % k)
+            else:
+                continue
+        if k in namesmap:
+            k = namesmap[k]
+        r[k] = v
+    return r
+
+
+def _namedtupleload(l: Loader, value: Dict[str, Any], type_) -> Any:
     """
     This loads a Dict[str, Any] into a NamedTuple.
     """
@@ -414,28 +469,37 @@ def _namedtupleload(l: Loader, value: Dict[str, Any], type_) -> Tuple:
         #dataclass
         import dataclasses
         fields = set(type_.__dataclass_fields__.keys())
-        optional_fields = {k for k,v in type_.__dataclass_fields__.items() if not (isinstance(getattr(v, 'default', dataclasses._MISSING_TYPE()), dataclasses._MISSING_TYPE) and isinstance(getattr(v, 'default_factory', dataclasses._MISSING_TYPE()), dataclasses._MISSING_TYPE))}
+        optional_fields = {k for k,v in type_.__dataclass_fields__.items() if
+                           v.init == False or
+                           not isinstance(v.default, dataclasses._MISSING_TYPE) or
+                           not isinstance(v.default_factory, dataclasses._MISSING_TYPE)}
         type_hints = {k: v.type for k,v in type_.__dataclass_fields__.items()}
 
         #Name mangling
 
         # Prepare the list of the needed name changes
-        transforms = []  # type: List[Tuple[str, str]]
-        for field in fields:
-            if type_.__dataclass_fields__[field].metadata:
-                name = type_.__dataclass_fields__[field].metadata.get('name')
+        transforms = {}  # type: Dict[str, str]
+        for pyname in fields:
+            if type_.__dataclass_fields__[pyname].metadata:
+                name = type_.__dataclass_fields__[pyname].metadata.get(l.mangle_key)
                 if name:
-                    transforms.append((field, name))
-        # Do the needed name changes
-        if transforms:
-            value = value.copy()
-            for pyname, dataname in transforms:
-                if dataname in value:
-                    tmp = value[dataname]
-                    del value[dataname]
-                    value[pyname] = tmp
+                    transforms[name] = pyname
+
+        try:
+            value = _mangle_names(transforms, value, l.failonextra)
+        except ValueError as e:
+            raise TypedloadValueError(str(e), value=value, type_=type_)
+
+    if hasattr(type_, '__required_keys__') and hasattr(type_, '__optional_keys__'):
+        # TypedDict, since 3.9
+        necessary_fields = type_.__required_keys__
+        optional_fields = type_.__optional_keys__
+    elif getattr(type_, '__total__', True) == False:
+        # TypedDict, only for 3.8
+        necessary_fields = set()
+    else:
+        necessary_fields = fields.difference(optional_fields)
 
-    necessary_fields = fields.difference(optional_fields)
     try:
         vfields = set(value.keys())
     except AttributeError as e:
@@ -445,7 +509,7 @@ def _namedtupleload(l: Loader, value: Dict[str, Any], type_) -> Tuple:
         raise TypedloadValueError(
             'Value does not contain fields: %s which are necessary for type %s' % (
                 necessary_fields.difference(vfields),
-                type_
+                tname(type_)
             ),
             value=value,
             type_=type_,
@@ -455,7 +519,7 @@ def _namedtupleload(l: Loader, value: Dict[str, Any], type_) -> Tuple:
     if l.failonextra and len(fieldsdiff):
         extra = ', '.join(fieldsdiff)
         raise TypedloadValueError(
-            'Dictionary has unrecognized fields: %s and cannot be loaded into %s' % (extra, type_),
+            'Dictionary has unrecognized fields: %s and cannot be loaded into %s' % (extra, tname(type_)),
             value=value,
             type_=type_,
         )
@@ -469,7 +533,12 @@ def _namedtupleload(l: Loader, value: Dict[str, Any], type_) -> Tuple:
             type_hints[k],
             annotation=Annotation(AnnotationType.FIELD, k),
         )
-    return type_(**params)
+    try:
+        return type_(**params)
+    except TypeError as e:
+        raise TypedloadTypeError(e)
+    except ValueError as e:
+        raise TypedloadValueError(e)
 
 
 def _unionload(l: Loader, value, type_) -> Any:
@@ -498,11 +567,7 @@ def _unionload(l: Loader, value, type_) -> Any:
     # Give a score to the types
     sorted_args = []  # type: List[Type]
     for t in args:
-        if (value_type == list and is_list(t)) or \
-           (value_type == dict and is_dict(t)) or \
-           (value_type == frozenset and is_frozenset(t)) or \
-           (value_type == set and is_set(t)) or \
-           (value_type == tuple and is_tuple(t)):
+        if t not in l.basictypes:
             sorted_args.insert(0, t)
         else:
             sorted_args.append(t)
@@ -514,7 +579,7 @@ def _unionload(l: Loader, value, type_) -> Any:
         except Exception as e:
             exceptions.append(e)
     raise TypedloadValueError(
-        'Value could not be loaded into %s' % type_,
+        'Value of %s could not be loaded into %s' % (tname(value_type), tname(type_)),
             value=value,
             type_=type_,
             exceptions=exceptions
@@ -546,8 +611,13 @@ def _enumload(l: Loader, value, type_) -> Enum:
             return type_(l.load(value, t, annotation=Annotation(AnnotationType.UNION, t)))
         except Exception as e:
             exceptions.append(e)
+    if len(type_.__members__) <= 10 and all(type(i.value) in l.basictypes for i in type_.__members__.values()):
+        lst = '\nValue %s not between: ' % repr(value) + \
+        ', '.join(repr(i.value) for i in type_.__members__.values())
+    else:
+        lst = ''
     raise TypedloadValueError(
-        'Value could not be loaded into %s' % type_,
+        'Value of %s could not be loaded into %s%s' % (tname(type(value)), tname(type_), lst),
         value=value,
         type_=type_,
         exceptions=exceptions
@@ -573,11 +643,12 @@ def _datetimeload(l: Loader, value, type_) -> Union[datetime.date, datetime.time
 
 def _attrload(l, value, type_):
     if not isinstance(value, dict):
-        raise TypedloadTypeError('Expected dictionary, got %s' % type(value), type_=type_, value=value)
+        raise TypedloadTypeError('Expected dictionary, got %s' % tname(type(value)), type_=type_, value=value)
     value = value.copy()
     names = []
     defaults = {}
     types = {}
+    namesmap = {}  # type: Dict[str, str]
 
     for attribute in type_.__attrs_attrs__:
         names.append(attribute.name)
@@ -585,14 +656,13 @@ def _attrload(l, value, type_):
         defaults[attribute.name] = attribute.default
 
         # Manage name mangling
-        if 'name' in attribute.metadata:
-            dataname = attribute.metadata['name']
-            pyname = attribute.name
+        if l.mangle_key in attribute.metadata:
+            namesmap[attribute.metadata[l.mangle_key]] = attribute.name
 
-            if dataname in value:
-                tmp = value[dataname]
-                del value[dataname]
-                value[pyname] = tmp
+    try:
+        value = _mangle_names(namesmap, value, l.failonextra)
+    except ValueError as e:
+        raise TypedloadValueError(str(e), value=value, type_=type_)
 
     t = _FakeNamedTuple((
         tuple(names),
@@ -604,8 +674,15 @@ def _attrload(l, value, type_):
     return _namedtupleload(l, value, t)
 
 
-def _pathload(l: Loader, value, type_) -> Path:
+def _strconstructload(l: Loader, value, type_):
+    """
+    Loader for all the types in the stdlib ipaddress module.
+    """
     try:
-        return Path(value)
+        return type_(value)
+    except ValueError as e:
+        raise TypedloadValueError(str(e), type_=type_, value=value)
     except TypeError as e:
         raise TypedloadTypeError(str(e), type_=type_, value=value)
+    except Exception as e:
+        raise TypedloadException(str(e), type_=type, value=value)
